@@ -85,9 +85,7 @@ TODO:
 #include <mysqld_error.h>
 #include <my_dir.h>
 #include <signal.h>
-#include <stdarg.h>
 #include <sslopt-vars.h>
-#include <sys/types.h>
 #ifndef __WIN__
 #include <sys/wait.h>
 #endif
@@ -146,7 +144,7 @@ static unsigned long connect_flags= CLIENT_MULTI_RESULTS |
                                     CLIENT_MULTI_STATEMENTS |
                                     CLIENT_REMEMBER_OPTIONS;
 
-static int verbose, delimiter_length;
+static int verbose;
 static uint commit_rate;
 static uint detach_rate;
 const char *num_int_cols_opt;
@@ -170,6 +168,7 @@ static ulonglong auto_generate_sql_number;
 const char *concurrency_str= NULL;
 static char *create_string;
 uint *concurrency;
+static char mysql_charsets_dir[FN_REFLEN+1];
 
 const char *default_dbug_option="d:t:o,/tmp/mysqlslap.trace";
 const char *opt_csv_str;
@@ -264,7 +263,7 @@ void option_cleanup(option_string *stmt);
 void concurrency_loop(MYSQL *mysql, uint current, option_string *eptr);
 static int run_statements(MYSQL *mysql, statement *stmt);
 int slap_connect(MYSQL *mysql);
-static int run_query(MYSQL *mysql, const char *query, int len);
+static int run_query(MYSQL *mysql, const char *query, size_t len);
 
 static const char ALPHANUMERICS[]=
   "0123456789ABCDEFGHIJKLMNOPQRSTWXYZabcdefghijklmnopqrstuvwxyz";
@@ -301,8 +300,12 @@ void set_mysql_connect_options(MYSQL *mysql)
     mysql_options(mysql,MYSQL_OPT_COMPRESS,NullS);
 #ifdef HAVE_OPENSSL
   if (opt_use_ssl)
+  {
     mysql_ssl_set(mysql, opt_ssl_key, opt_ssl_cert, opt_ssl_ca,
                   opt_ssl_capath, opt_ssl_cipher);
+    mysql_options(mysql, MYSQL_OPT_SSL_CRL, opt_ssl_crl);
+    mysql_options(mysql, MYSQL_OPT_SSL_CRLPATH, opt_ssl_crlpath);
+  }
 #endif
   if (opt_protocol)
     mysql_options(mysql,MYSQL_OPT_PROTOCOL,(char*)&opt_protocol);
@@ -322,11 +325,7 @@ int main(int argc, char **argv)
   MY_INIT(argv[0]);
   sf_leaking_memory=1; /* don't report memory leaks on early exits */
 
-  if (load_defaults("my",load_default_groups,&argc,&argv))
-  {
-    my_end(0);
-    exit(1);
-  }
+  load_defaults_or_exit("my", load_default_groups, &argc, &argv);
   defaults_argv=argv;
   if (get_options(&argc,&argv))
   {
@@ -340,9 +339,6 @@ int main(int argc, char **argv)
   if (auto_generate_sql)
     srandom((uint)time(NULL));
 
-  /* globals? Yes, so we only have to run strlen once */
-  delimiter_length= strlen(delimiter);
-
   if (argc > 2)
   {
     fprintf(stderr,"%s: Too many arguments\n",my_progname);
@@ -359,6 +355,9 @@ int main(int argc, char **argv)
   if (opt_default_auth && *opt_default_auth)
     mysql_options(&mysql, MYSQL_DEFAULT_AUTH, opt_default_auth);
 
+  mysql_options(&mysql, MYSQL_OPT_CONNECT_ATTR_RESET, 0);
+  mysql_options4(&mysql, MYSQL_OPT_CONNECT_ATTR_ADD,
+                 "program_name", "mysqlslap");
   if (!opt_only_print) 
   {
     if (!(mysql_real_connect(&mysql, host, user, opt_password,
@@ -367,6 +366,7 @@ int main(int argc, char **argv)
     {
       fprintf(stderr,"%s: Error when connecting to server: %s\n",
               my_progname,mysql_error(&mysql));
+      mysql_close(&mysql);
       free_defaults(defaults_argv);
       my_end(0);
       exit(1);
@@ -412,8 +412,7 @@ int main(int argc, char **argv)
   pthread_mutex_destroy(&sleeper_mutex);
   pthread_cond_destroy(&sleep_threshhold);
 
-  if (!opt_only_print) 
-    mysql_close(&mysql); /* Close & free connection */
+  mysql_close(&mysql); /* Close & free connection */
 
   /* now free all the strings we created */
   my_free(opt_password);
@@ -580,6 +579,9 @@ static struct my_option my_long_options[] =
     "Number of row inserts to perform for each thread (default is 100).",
     &auto_generate_sql_number, &auto_generate_sql_number,
     0, GET_ULL, REQUIRED_ARG, 100, 0, 0, 0, 0, 0},
+  {"character-sets-dir", OPT_CHARSETS_DIR,
+   "Directory for character set files.", (char **)&charsets_dir,
+   (char **)&charsets_dir, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"commit", OPT_SLAP_COMMIT, "Commit records every X number of statements.",
     &commit_rate, &commit_rate, 0, GET_UINT, REQUIRED_ARG,
     0, 0, 0, 0, 0, 0},
@@ -770,13 +772,21 @@ get_one_option(int optid, const struct my_option *opt __attribute__((unused)),
 #endif
     break;
   case OPT_MYSQL_PROTOCOL:
-    opt_protocol= find_type_or_exit(argument, &sql_protocol_typelib,
-                                    opt->name);
+    if ((opt_protocol= find_type_with_warning(argument, &sql_protocol_typelib,
+                                              opt->name)) <= 0)
+    {
+      sf_leaking_memory= 1; /* no memory leak reports here */
+      exit(1);
+    }
     break;
   case '#':
     DBUG_PUSH(argument ? argument : default_dbug_option);
     debug_check_flag= 1;
     break;
+  case OPT_CHARSETS_DIR:
+    strmake_buf(mysql_charsets_dir, argument);
+    charsets_dir = mysql_charsets_dir;
+    break;
   case OPT_SLAP_CSV:
     if (!argument)
       argument= (char *)"-"; /* use stdout */
@@ -804,7 +814,7 @@ get_random_string(char *buf)
   DBUG_ENTER("get_random_string");
   for (x= RAND_STRING_SIZE; x > 0; x--)
     *buf_ptr++= ALPHANUMERICS[random() % ALPHANUMERICS_SIZE];
-  DBUG_RETURN(buf_ptr - buf);
+  DBUG_RETURN((uint)(buf_ptr - buf));
 }
 
 
@@ -1183,9 +1193,6 @@ get_options(int *argc,char ***argv)
   if (debug_check_flag)
     my_end_arg= MY_CHECK_ERROR;
 
-  if (!user)
-    user= (char *)"root";
-
   /*
     If something is created and --no-drop is not specified, we drop the
     schema.
@@ -1535,18 +1542,18 @@ get_options(int *argc,char ***argv)
 }
 
 
-static int run_query(MYSQL *mysql, const char *query, int len)
+static int run_query(MYSQL *mysql, const char *query, size_t len)
 {
   if (opt_only_print)
   {
-    printf("%.*s;\n", len, query);
+    printf("%.*s;\n", (int)len, query);
     return 0;
   }
 
   if (verbose >= 3)
-    printf("%.*s;\n", len, query);
+    printf("%.*s;\n", (int)len, query);
 
-  return mysql_real_query(mysql, query, len);
+  return mysql_real_query(mysql, query, (ulong)len);
 }
 
 
@@ -1808,8 +1815,8 @@ run_scheduler(stats *sptr, statement *stmts, uint concur, ulonglong limit)
 
   pthread_mutex_lock(&sleeper_mutex);
   master_wakeup= 0;
-  pthread_mutex_unlock(&sleeper_mutex);
   pthread_cond_broadcast(&sleep_threshhold);
+  pthread_mutex_unlock(&sleeper_mutex);
 
   gettimeofday(&start_time, NULL);
 
@@ -1858,21 +1865,21 @@ pthread_handler_t run_task(void *p)
   }
   pthread_mutex_unlock(&sleeper_mutex);
 
-  if (!(mysql= mysql_init(NULL)))
-  {
-    fprintf(stderr,"%s: mysql_init() failed ERROR : %s\n",
-            my_progname, mysql_error(mysql));
-    exit(0);
-  }
-  set_mysql_connect_options(mysql);
-
   if (mysql_thread_init())
   {
-    fprintf(stderr,"%s: mysql_thread_init() failed ERROR : %s\n",
-            my_progname, mysql_error(mysql));
+    fprintf(stderr,"%s: mysql_thread_init() failed\n", my_progname);
     exit(0);
   }
 
+  if (!(mysql= mysql_init(NULL)))
+  {
+    fprintf(stderr,"%s: mysql_init() failed\n", my_progname);
+    mysql_thread_end();
+    exit(0);
+  }
+
+  set_mysql_connect_options(mysql);
+
   DBUG_PRINT("info", ("trying to connect to host %s as user %s", host, user));
 
   if (!opt_only_print)
@@ -1990,8 +1997,7 @@ end:
   if (commit_rate)
     run_query(mysql, "COMMIT", strlen("COMMIT"));
 
-  if (!opt_only_print) 
-    mysql_close(mysql);
+  mysql_close(mysql);
 
   mysql_thread_end();
 
@@ -2115,7 +2121,7 @@ parse_delimiter(const char *script, statement **stmt, char delm)
   char *ptr= (char *)script;
   statement **sptr= stmt;
   statement *tmp;
-  uint length= strlen(script);
+  size_t length= strlen(script);
   uint count= 0; /* We know that there is always one */
 
   for (tmp= *sptr= (statement *)my_malloc(sizeof(statement),
@@ -2290,7 +2296,7 @@ slap_connect(MYSQL *mysql)
                            opt_mysql_unix_port,
                            connect_flags))
     {
-      /* Connect suceeded */
+      /* Connect succeeded */
       connect_error= 0;
       break;
     }
