@@ -23,36 +23,12 @@
 #include <m_string.h>
 #include <errno.h>
 #include <my_getopt.h>
-#ifdef WITH_NDBCLUSTER_STORAGE_ENGINE
-#include "../storage/ndb/src/ndbapi/ndberror.c"
-#include "../storage/ndb/src/kernel/error/ndbd_exit_codes.c"
-#include "../storage/ndb/include/mgmapi/mgmapi_error.h"
-#endif
 #include <welcome_copyright_notice.h> /* ORACLE_WELCOME_COPYRIGHT_NOTICE */
 
 static my_bool verbose, print_all_codes;
 
-#include "../include/my_base.h"
-#include "../mysys/my_handler_errors.h"
-
-#ifdef WITH_NDBCLUSTER_STORAGE_ENGINE
-static my_bool ndb_code;
-static char ndb_string[1024];
-int mgmapi_error_string(int err_no, char *str, int size)
-{
-  int i;
-  for (i= 0; i < ndb_mgm_noOfErrorMsgs; i++)
-  {
-    if ((int)ndb_mgm_error_msgs[i].code == err_no)
-    {
-      my_snprintf(str, size-1, "%s", ndb_mgm_error_msgs[i].msg);
-      str[size-1]= '\0';
-      return 0;
-    }
-  }
-  return -1;
-}
-#endif
+#include <my_base.h>
+#include <my_handler_errors.h>
 
 static struct my_option my_long_options[] =
 {
@@ -60,10 +36,6 @@ static struct my_option my_long_options[] =
    NO_ARG, 0, 0, 0, 0, 0, 0},
   {"info", 'I', "Synonym for --help.",  0, 0, 0, GET_NO_ARG,
    NO_ARG, 0, 0, 0, 0, 0, 0},
-#ifdef WITH_NDBCLUSTER_STORAGE_ENGINE
-  {"ndb", 257, "Ndbcluster storage engine specific error codes.", &ndb_code,
-   &ndb_code, 0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
-#endif
 #ifdef HAVE_SYS_ERRLIST
   {"all", 'a', "Print all the error messages and the number. Deprecated,"
    " will be removed in a future release.",
@@ -133,11 +105,13 @@ get_one_option(int optid, const struct my_option *opt __attribute__((unused)),
     break;
   case 'V':
     print_version();
+    my_end(0);
     exit(0);
     break;
   case 'I':
   case '?':
     usage();
+    my_end(0);
     exit(0);
     break;
   }
@@ -150,7 +124,10 @@ static int get_options(int *argc,char ***argv)
   int ho_error;
 
   if ((ho_error=handle_options(argc, argv, my_long_options, get_one_option)))
+  {
+    my_end(0);
     exit(ho_error);
+  }
 
   if (!*argc && !print_all_codes)
   {
@@ -232,7 +209,7 @@ static my_bool print_win_error_msg(DWORD error, my_bool verbose)
                     NULL))
   {
     if (verbose)
-      printf("Win32 error code %d: %s", error, s);
+      printf("Win32 error code %lu: %s", error, s);
     else
       puts(s);
     LocalFree(s);
@@ -250,7 +227,7 @@ static my_bool print_win_error_msg(DWORD error, my_bool verbose)
     will ignore calls to register already registered error numbers.
 */
 
-static const char **get_handler_error_messages()
+static const char **get_handler_error_messages(int e __attribute__((unused)))
 {
   return handler_error_messages;
 }
@@ -288,7 +265,10 @@ int main(int argc,char *argv[])
   MY_INIT(argv[0]);
 
   if (get_options(&argc,&argv))
+  {
+    my_end(0);
     exit(1);
+  }
 
   my_handler_error_register();
 
@@ -334,35 +314,7 @@ int main(int argc,char *argv[])
 
       found=0;
       code=atoi(*argv);
-#ifdef WITH_NDBCLUSTER_STORAGE_ENGINE
-      if (ndb_code)
-      {
-        if ((ndb_error_string(code, ndb_string, sizeof(ndb_string)) < 0) &&
-            (ndbd_exit_string(code, ndb_string, sizeof(ndb_string)) < 0) &&
-            (mgmapi_error_string(code, ndb_string, sizeof(ndb_string)) < 0))
-	{
-          msg= 0;
-	}
-	else
-	  msg= ndb_string;
-        if (msg)
-        {
-          if (verbose)
-            printf("NDB error code %3d: %s\n",code,msg);
-          else
-            puts(msg);
-        }
-        else
-        {
-	  fprintf(stderr,"Illegal ndb error code: %d\n",code);
-          error= 1;
-        }
-        found= 1;
-        msg= 0;
-      }
-      else
-#endif
-	msg = strerror(code);
+      msg = strerror(code);
 
       /*
         We don't print the OS error message if it is the same as the
