@@ -12,7 +12,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
 #include <my_global.h>
 #include <my_stacktrace.h>
@@ -34,107 +34,11 @@
 #include <execinfo.h>
 #endif
 
-#define PTR_SANE(p) ((p) && (char*)(p) >= heap_start && (char*)(p) <= heap_end)
-
-static char *heap_start;
-
-#ifdef HAVE_BSS_START
-extern char *__bss_start;
-#endif
-
-void my_init_stacktrace()
-{
-#ifdef HAVE_BSS_START
-  heap_start = (char*) &__bss_start;
-#endif
-}
-
-#ifdef __linux__
-
-static void print_buffer(char *buffer, size_t count)
-{
-  const char s[]= " ";
-  for (; count && *buffer; --count)
-  {
-    my_write_stderr(isprint(*buffer) ? buffer : s, 1);
-    ++buffer;
-  }
-}
-
-/**
-  Access the pages of this process through /proc/self/task/<tid>/mem
-  in order to safely print the contents of a memory address range.
-
-  @param  addr      The address at the start of the memory region.
-  @param  max_len   The length of the memory region.
-
-  @return Zero on success.
-*/
-static int safe_print_str(const char *addr, int max_len)
-{
-  int fd;
-  pid_t tid;
-  off_t offset;
-  ssize_t nbytes= 0;
-  size_t total, count;
-  char buf[256];
-
-  tid= (pid_t) syscall(SYS_gettid);
-
-  sprintf(buf, "/proc/self/task/%d/mem", tid);
-
-  if ((fd= open(buf, O_RDONLY)) < 0)
-    return -1;
-
-  /* Ensure that off_t can hold a pointer. */
-  compile_time_assert(sizeof(off_t) >= sizeof(intptr));
-
-  total= max_len;
-  offset= (intptr) addr;
-
-  /* Read up to the maximum number of bytes. */
-  while (total)
-  {
-    count= MY_MIN(sizeof(buf), total);
-
-    if ((nbytes= pread(fd, buf, count, offset)) < 0)
-    {
-      /* Just in case... */
-      if (errno == EINTR)
-        continue;
-      else
-        break;
-    }
-
-    /* Advance offset into memory. */
-    total-= nbytes;
-    offset+= nbytes;
-    addr+= nbytes;
-
-    /* Output the printable characters. */
-    print_buffer(buf, nbytes);
-
-    /* Break if less than requested... */
-    if ((count - nbytes))
-      break;
-  }
-
-  if (nbytes == -1)
-    my_safe_printf_stderr("Can't read from address %p", addr);
-
-  close(fd);
-
-  return 0;
-}
-
-#endif
-
 /*
   Attempt to print a char * pointer as a string.
 
   SYNOPSIS
-    Prints either until the end of string ('\0'), or max_len characters have
-    been printed.
+    Prints until  max_len characters have been printed.
 
   RETURN VALUE
     0  Pointer was within the heap address space.
@@ -147,26 +51,27 @@ static int safe_print_str(const char *addr, int max_len)
     returns 1, it does not mean 100% that the pointer is corrupted.
 */
 
-int my_safe_print_str(const char* val, int max_len)
+int my_safe_print_str(const char* val, size_t max_len)
 {
-  char *heap_end;
-
-#ifdef __linux__
-  // Try and make use of /proc filesystem to safely print memory contents.
-  if (!safe_print_str(val, max_len))
-    return 0;
-#endif
-
-  heap_end= (char*) sbrk(0);
-
-  if (!PTR_SANE(val))
+  const char *orig_val= val;
+  if (!val)
   {
-    my_safe_printf_stderr("%s", "is an invalid pointer");
+    my_safe_printf_stderr("%s", "(null)");
     return 1;
   }
 
-  for (; max_len && PTR_SANE(val) && *val; --max_len)
-    my_write_stderr((val++), 1);
+  for (; max_len; --max_len)
+  {
+    if (my_write_stderr((val++), 1) != 1)
+    {
+      if ((errno == EFAULT) &&(val == orig_val + 1))
+      {
+        // We can not read the address from very beginning
+        my_safe_printf_stderr("Can't access address %p", orig_val);
+      }
+      break;
+    }
+  }
   my_safe_printf_stderr("%s", "\n");
 
   return 0;
@@ -178,12 +83,13 @@ int my_safe_print_str(const char* val, int max_len)
 #include <ucontext.h>
 
 void my_print_stacktrace(uchar* stack_bottom __attribute__((unused)), 
-                         ulong thread_stack __attribute__((unused)))
+                         ulong thread_stack __attribute__((unused)),
+                         my_bool silent)
 {
   if (printstack(fileno(stderr)) == -1)
     my_safe_printf_stderr("%s",
       "Error when traversing the stack, stack appears corrupt.\n");
-  else
+  else if (!silent)
     my_safe_printf_stderr("%s",
       "Please read "
       "http://dev.mysql.com/doc/refman/5.1/en/resolve-stack-dump.html\n"
@@ -260,7 +166,8 @@ static int print_with_addr_resolve(void **addrs, int n)
 }
 #endif
 
-void my_print_stacktrace(uchar* stack_bottom, ulong thread_stack)
+void my_print_stacktrace(uchar* stack_bottom, ulong thread_stack,
+                         my_bool silent __attribute__((unused)))
 {
   void *addrs[128];
   char **strings __attribute__((unused)) = NULL;
@@ -334,14 +241,14 @@ inline uint32* find_prev_pc(uint32* pc, uchar** fp)
 }
 #endif /* defined(__alpha__) && defined(__GNUC__) */
 
-void my_print_stacktrace(uchar* stack_bottom, ulong thread_stack)
+void my_print_stacktrace(uchar* stack_bottom, ulong thread_stack,
+                         my_bool silent)
 {
-  uchar** fp;
+  uchar** UNINIT_VAR(fp);
   uint frame_count = 0, sigreturn_frame_count;
 #if defined(__alpha__) && defined(__GNUC__)
   uint32* pc;
 #endif
-  LINT_INIT(fp);
 
 
 #ifdef __i386__
@@ -450,7 +357,8 @@ void my_print_stacktrace(uchar* stack_bottom, ulong thread_stack)
                         "Stack trace seems successful - bottom reached\n");
 
 end:
-  my_safe_printf_stderr("%s",
+  if (!silent)
+    my_safe_printf_stderr("%s",
     "Please read "
     "http://dev.mysql.com/doc/refman/5.1/en/resolve-stack-dump.html\n"
     "and follow instructions on how to resolve the stack trace.\n"
@@ -484,7 +392,18 @@ void my_write_core(int sig)
 
 #else /* __WIN__*/
 
+#ifdef _MSC_VER
+/* Silence warning in OS header dbghelp.h */
+#pragma warning(push)
+#pragma warning(disable : 4091)
+#endif
+
 #include <dbghelp.h>
+
+#ifdef _MSC_VER
+#pragma warning(pop)
+#endif
+
 #include <tlhelp32.h>
 #include <my_sys.h>
 #if _MSC_VER
@@ -496,11 +415,6 @@ static EXCEPTION_POINTERS *exception_ptrs;
 #define MODULE64_SIZE_WINXP 576
 #define STACKWALK_MAX_FRAMES 64
 
-void my_init_stacktrace()
-{
-}
-
-
 void my_set_exception_pointers(EXCEPTION_POINTERS *ep)
 {
   exception_ptrs = ep;
@@ -600,7 +514,7 @@ static void get_symbol_path(char *path, size_t size)
 #define SYMOPT_NO_PROMPTS 0
 #endif
 
-void my_print_stacktrace(uchar* unused1, ulong unused2)
+void my_print_stacktrace(uchar* unused1, ulong unused2, my_bool silent)
 {
   HANDLE  hProcess= GetCurrentProcess();
   HANDLE  hThread= GetCurrentThread();
@@ -689,7 +603,7 @@ void my_print_stacktrace(uchar* unused1, ulong unused2)
     if(have_source)
     {
       const char *base_file_name= my_basename(line.FileName);
-      my_safe_printf_stderr("[%s:%u]",
+      my_safe_printf_stderr("[%s:%lu]",
                             base_file_name, line.LineNumber);
     }
     my_safe_printf_stderr("%s", "\n");
@@ -719,7 +633,7 @@ void my_write_core(int unused)
   if(GetModuleFileName(NULL, path, sizeof(path)))
   {
     _splitpath(path, NULL, NULL,dump_fname,NULL);
-    strncat(dump_fname, ".dmp", sizeof(dump_fname));
+    strcat_s(dump_fname, sizeof(dump_fname), ".dmp");
   }
 
   hFile= CreateFile(dump_fname, GENERIC_WRITE, 0, 0, CREATE_ALWAYS,
@@ -749,7 +663,7 @@ void my_write_core(int unused)
 }
 
 
-int my_safe_print_str(const char *val, int len)
+int my_safe_print_str(const char *val, size_t len)
 {
   __try
   {
@@ -766,7 +680,7 @@ int my_safe_print_str(const char *val, int len)
 
 size_t my_write_stderr(const void *buf, size_t count)
 {
-  return (size_t) write(fileno(stderr), buf, count);
+  return (size_t) write(fileno(stderr), buf, (uint)count);
 }
 
 
