@@ -20,7 +20,8 @@ mysqld_ld_preload=
 mysqld_ld_library_path=
 flush_caches=0
 numa_interleave=0
-unsafe_my_cnf=0
+wsrep_on=0
+dry_run=0
 defaults_group_suffix=
 
 # Initial logging status: error log is not open, and not using syslog
@@ -28,9 +29,11 @@ logging=init
 want_syslog=0
 syslog_tag=
 user='@MYSQLD_USER@'
+group='@MYSQLD_GROUP@'
 pid_file=
 err_log=
 err_log_base=
+skip_err_log=0
 
 syslog_tag_mysqld=mysqld
 syslog_tag_mysqld_safe=mysqld_safe
@@ -71,16 +74,19 @@ usage () {
         cat <<EOF
 Usage: $0 [OPTIONS]
   --no-defaults              Don't read the system defaults file
-  --core-file-size=LIMIT     Limit core files to the specified size
   --defaults-file=FILE       Use the specified defaults file
   --defaults-extra-file=FILE Also use defaults from the specified file
+  --defaults-group-suffix=X  Additionally read default groups with X appended
+                             as a suffix
   --ledir=DIRECTORY          Look for mysqld in the specified directory
   --open-files-limit=LIMIT   Limit the number of open files
   --crash-script=FILE        Script to call when mysqld crashes
+  --core-file-size=LIMIT     Limit core files to the specified size
   --timezone=TZ              Set the system timezone
   --malloc-lib=LIB           Preload shared library LIB if available
   --mysqld=FILE              Use the specified file as mysqld
   --mysqld-version=VERSION   Use "mysqld-VERSION" as mysqld
+  --dry-run                  Simulate the start to detect errors but don't start
   --nice=NICE                Set the scheduling priority of mysqld
   --no-auto-restart          Exit after starting mysqld
   --nowatch                  Exit after starting mysqld
@@ -101,35 +107,6 @@ EOF
         exit 1
 }
 
-my_which ()
-{
-  save_ifs="${IFS-UNSET}"
-  IFS=:
-  ret=0
-  for file
-  do
-    for dir in $PATH
-    do
-      if [ -f "$dir/$file" ]
-      then
-        echo "$dir/$file"
-        continue 2
-      fi
-    done
-
-	ret=1  #signal an error
-	break
-  done
-
-  if [ "$save_ifs" = UNSET ]
-  then
-    unset IFS
-  else
-    IFS="$save_ifs"
-  fi
-  return $ret  # Success
-}
-
 find_in_bin() {
   if test -x "$MY_BASEDIR_VERSION/bin/$1"
   then
@@ -143,6 +120,7 @@ find_in_bin() {
 }
 
 log_generic () {
+  [ $dry_run -eq 1 ] && return
   priority="$1"
   shift
 
@@ -150,7 +128,11 @@ log_generic () {
   echo "$msg"
   case $logging in
     init) ;;  # Just echo the message, don't save it anywhere
-    file) echo "$msg" | "$helper" "$user" log "$err_log" ;;
+    file)
+      if [ "$helper_exist" -eq "0" ]; then
+        echo "$msg" | "$helper" "$user" log "$err_log"
+      fi
+    ;;
     syslog) logger -t "$syslog_tag_mysqld_safe" -p "$priority" "$*" ;;
     *)
       echo "Internal program error (non-fatal):" \
@@ -168,9 +150,13 @@ log_notice () {
 }
 
 eval_log_error () {
-  cmd="$1"
+  local cmd="$1"
   case $logging in
-    file) cmd="$cmd 2>&1 | "`shell_quote_string "$helper"`" $user log "`shell_quote_string "$err_log"` ;;
+    file)
+     if [ "$helper_exist" -eq "0" ]; then
+        cmd="$cmd 2>&1 | "`shell_quote_string "$helper"`" $user log "`shell_quote_string "$err_log"`
+     fi
+     ;;
     syslog)
       # mysqld often prefixes its messages with a timestamp, which is
       # redundant when logging to syslog (which adds its own timestamp)
@@ -204,67 +190,176 @@ shell_quote_string() {
   echo "$1" | sed -e 's,\([^a-zA-Z0-9/_.=-]\),\\\1,g'
 }
 
-check_executable_location() {
-  if test "$unsafe_my_cnf" = 1 -a "$unrecognized_handling" != collect; then
-    log_error "Cannot accept $1 from a config file, when my.cnf is in the datadir"
+wsrep_pick_url() {
+  [ $# -eq 0 ] && return 0
+
+  log_error "WSREP: 'wsrep_urls' is DEPRECATED! Use wsrep_cluster_address to specify multiple addresses instead."
+
+  if ! command -v nc >/dev/null
+  then
+    log_error "ERROR: nc tool not found in PATH! Make sure you have it installed."
+    return 1
+  fi
+
+  local url
+  # Assuming URL in the form scheme://host:port
+  # If host and port are not NULL, the liveness of URL is assumed to be tested
+  # If port part is absent, the url is returned literally and unconditionally
+  # If every URL has port but none is reachable, nothing is returned
+  for url in `echo $@ | sed s/,/\ /g` 0; do
+    local host=`echo $url | cut -d \: -f 2 | sed s/^\\\/\\\///`
+    local port=`echo $url | cut -d \: -f 3`
+    [ -z "$port" ] && break
+    nc -z "$host" $port >/dev/null && break
+  done
+
+  if [ "$url" == "0" ]; then
+    log_error "ERROR: none of the URLs in '$@' is reachable."
+    return 1
+  fi
+
+  echo $url
+}
+
+# Run mysqld with --wsrep-recover and parse recovered position from log.
+# Position will be stored in wsrep_start_position_opt global.
+wsrep_start_position_opt=""
+wsrep_recover_position() {
+  local mysqld_cmd="$@"
+  local euid=$(id -u)
+  local ret=0
+
+  local wr_logfile=$(mktemp /tmp/wsrep_recovery.XXXXXX)
+
+  # safety checks
+  if [ -z $wr_logfile ]; then
+    log_error "WSREP: mktemp failed"
+    return 1
+  fi
+
+  if [ -f $wr_logfile ]; then
+    # NOTE! Do not change ownership of the temporary file, as on newer kernel
+    # versions fs.protected_regular is set to '2' and redirecting output with >
+    # as root to a file not owned by root will fail with "Permission denied"
+    chmod 600 $wr_logfile
+  else
+    log_error "WSREP: mktemp failed"
+    return 1
+  fi
+
+  local wr_pidfile="$DATADIR/"`@HOSTNAME@`"-recover.pid"
+
+  local wr_options="--disable-log-error  --pid-file='$wr_pidfile'"
+
+  log_notice "WSREP: Running position recovery with $wr_options"
+
+  eval "$mysqld_cmd --wsrep_recover $wr_options 2> $wr_logfile"
+
+  if [ ! -s "$wr_logfile" ]; then
+    log_error "Log file $wr_logfile was empty, cannot proceed. Is system running fs.protected_regular?"
     exit 1
   fi
+
+  local rp="$(grep 'WSREP: Recovered position:' $wr_logfile)"
+  if [ -z "$rp" ]; then
+    local skipped="$(grep WSREP $wr_logfile | grep 'skipping position recovery')"
+    if [ -z "$skipped" ]; then
+      log_error "WSREP: Failed to recover position: '`cat $wr_logfile`'"
+      ret=1
+    else
+      log_notice "WSREP: Position recovery skipped"
+    fi
+  else
+    local start_pos="$(echo $rp | sed 's/.*WSREP\:\ Recovered\ position://' \
+        | sed 's/^[ \t]*//')"
+    log_notice "WSREP: Recovered position $start_pos"
+    wsrep_start_position_opt="--wsrep_start_position=$start_pos"
+  fi
+
+  [ $ret -eq 0 ] && rm $wr_logfile
+
+  return $ret
 }
 
 parse_arguments() {
   for arg do
-    # the parameter after "=", or the whole $arg if no match
-    val=`echo "$arg" | sed -e 's;^--[^=]*=;;'`
-    # what's before "=", or the whole $arg if no match
-    optname=`echo "$arg" | sed -e 's/^\(--[^=]*\)=.*$/\1/'`
-    # replace "_" by "-" ; mysqld_safe must accept "_" like mysqld does.
-    optname_subst=`echo "$optname" | sed 's/_/-/g'`
-    arg=`echo $arg | sed "s/^$optname/$optname_subst/"`
+    val=`echo "$arg" | sed -e "s;--[^=]*=;;"`
     case "$arg" in
       # these get passed explicitly to mysqld
       --basedir=*) MY_BASEDIR_VERSION="$val" ;;
       --datadir=*|--data=*) DATADIR="$val" ;;
-      --pid-file=*) pid_file="$val" ;;
-      --plugin-dir=*) PLUGIN_DIR="$val" ;;
+      --pid[-_]file=*) pid_file="$val" ;;
+      --plugin[-_]dir=*) PLUGIN_DIR="$val" ;;
       --user=*) user="$val"; SET_USER=1 ;;
-      --log-basename=*|--hostname=*|--loose-log-basename=*)
+      --group=*) group="$val"; SET_USER=1 ;;
+      --log[-_]basename=*|--hostname=*|--loose[-_]log[-_]basename=*)
         pid_file="$val.pid";
 	err_log_base="$val";
 	;;
 
       # these might have been set in a [mysqld_safe] section of my.cnf
       # they are added to mysqld command line to override settings from my.cnf
-      --log-error=*) err_log="$val" ;;
+      --skip[-_]log[-_]error)
+        err_log=;
+        skip_err_log=1;
+        ;;
+      --log[-_]error=*)
+        err_log="$val";
+        skip_err_log=0;
+        ;;
       --port=*) mysql_tcp_port="$val" ;;
       --socket=*) mysql_unix_port="$val" ;;
 
       # mysqld_safe-specific options - must be set in my.cnf ([mysqld_safe])!
-      --core-file-size=*) core_file_size="$val" ;;
-      --ledir=*) check_executable_location "$arg" ; ledir="$val" ;;
-      --malloc-lib=*) check_executable_location "$arg"; set_malloc_lib "$val" ;;
-      --crash-script=*) check_executable_location "$arg"; crash_script="$val" ;;
-      --mysqld=*) check_executable_location "$arg"; MYSQLD="$val" ;;
-      --mysqld-version=*)
+      --core[-_]file[-_]size=*) core_file_size="$val" ;;
+      --ledir=*) ledir="$val" ;;
+      --malloc[-_]lib=*) set_malloc_lib "$val" ;;
+      --crash[-_]script=*) crash_script="$val" ;;
+      --mysqld=*) MYSQLD="$val" ;;
+      --mysqld[-_]version=*)
         if test -n "$val"
         then
-          check_executable_location "$arg"
           MYSQLD="mysqld-$val"
           PLUGIN_VARIANT="/$val"
         else
           MYSQLD="mysqld"
         fi
         ;;
+      --dry[-_]run) dry_run=1 ;;
       --nice=*) niceness="$val" ;;
-      --nowatch|--no-watch|--no-auto-restart) nowatch=1 ;;
-      --open-files-limit=*) open_files="$val" ;;
-      --open_files_limit=*) open_files="$val" ;;
-      --skip-kill-mysqld*) KILL_MYSQLD=0 ;;
+      --nowatch|--no[-_]watch|--no[-_]auto[-_]restart) nowatch=1 ;;
+      --open[-_]files[-_]limit=*) open_files="$val" ;;
+      --skip[-_]kill[-_]mysqld*) KILL_MYSQLD=0 ;;
       --syslog) want_syslog=1 ;;
-      --skip-syslog) want_syslog=0 ;;
-      --syslog-tag=*) syslog_tag="$val" ;;
+      --skip[-_]syslog) want_syslog=0 ;;
+      --syslog[-_]tag=*) syslog_tag="$val" ;;
       --timezone=*) TZ="$val"; export TZ; ;;
-      --flush-caches) flush_caches=1 ;;
-      --numa-interleave) numa_interleave=1 ;;
+      --flush[-_]caches) flush_caches=1 ;;
+      --numa[-_]interleave) numa_interleave=1 ;;
+      --wsrep[-_]on)
+        wsrep_on=1
+        append_arg_to_args "$arg"
+        ;;
+      --skip[-_]wsrep[-_]on)
+        wsrep_on=0
+        append_arg_to_args "$arg"
+        ;;
+      --wsrep[-_]on=*)
+        if echo $val | grep -iq '\(ON\|1\)'; then
+          wsrep_on=1
+        else
+          wsrep_on=0
+        fi
+        append_arg_to_args "$arg"
+        ;;
+      --wsrep[-_]urls=*) wsrep_urls="$val"; ;;
+      --wsrep[-_]provider=*)
+        if test -n "$val" && test "$val" != "none"
+        then
+          wsrep_restart=1
+        fi
+        append_arg_to_args "$arg"
+        ;;
 
       --defaults-group-suffix=*) defaults_group_suffix="$arg" ;;
 
@@ -337,53 +432,40 @@ mysqld_ld_preload_text() {
   echo "$text"
 }
 
-
-mysql_config=
-get_mysql_config() {
-  if [ -z "$mysql_config" ]; then
-    mysql_config=`echo "$0" | sed 's,/[^/][^/]*$,/mysql_config,'`
-    if [ ! -x "$mysql_config" ]; then
-      log_error "Can not run mysql_config $@ from '$mysql_config'"
-      exit 1
-    fi
-  fi
-
-  "$mysql_config" "$@"
-}
-
-
 # set_malloc_lib LIB
 # - If LIB is empty, do nothing and return
-# - If LIB is 'tcmalloc', look for tcmalloc shared library in /usr/lib
-#   then pkglibdir.  tcmalloc is part of the Google perftools project.
+# - If LIB starts with 'tcmalloc' or 'jemalloc', look for the shared library
+#   using `ldconfig`.
+#   tcmalloc is part of the Google perftools project.
 # - If LIB is an absolute path, assume it is a malloc shared library
 #
 # Put LIB in mysqld_ld_preload, which will be added to LD_PRELOAD when
 # running mysqld.  See ld.so for details.
 set_malloc_lib() {
   malloc_lib="$1"
-
-  if [ "$malloc_lib" = tcmalloc ]; then
-    pkglibdir=`get_mysql_config --variable=pkglibdir`
-    malloc_lib=
-    # This list is kept intentionally simple.  Simply set --malloc-lib
-    # to a full path if another location is desired.
-    for libdir in /usr/lib "$pkglibdir" "$pkglibdir/mysql"; do
-      for flavor in _minimal '' _and_profiler _debug; do
-        tmp="$libdir/libtcmalloc$flavor.so"
-        #log_notice "DEBUG: Checking for malloc lib '$tmp'"
-        [ -r "$tmp" ] || continue
-        malloc_lib="$tmp"
-        break 2
-      done
-    done
-
-    if [ -z "$malloc_lib" ]; then
-      log_error "no shared library for --malloc-lib=tcmalloc found in /usr/lib or $pkglibdir"
+  if expr "$malloc_lib" : "\(tcmalloc\|jemalloc\)" > /dev/null ; then
+    export PATH=$PATH:/sbin
+    if ! command -v ldconfig > /dev/null 2>&1
+    then
+      log_error "ldconfig command not found, required for ldconfig -p"
       exit 1
     fi
-  fi
+    # format from ldconfig:
+    # "libjemalloc.so.1 (libc6,x86-64) => /usr/lib/x86_64-linux-gnu/libjemalloc.so.1"
+    libmalloc_path="$(ldconfig -p | sed -n "/lib${malloc_lib}/p" | cut -d '>' -f2)"
 
+    if [ -z "$libmalloc_path" ]; then
+      log_error "no shared library for lib$malloc_lib.so.[0-9] found."
+      exit 1
+    fi
+
+    for f in $libmalloc_path; do
+      if [ -f "$f" ]; then
+        malloc_lib=$f # get the first path if many
+        break
+      fi
+    done
+  fi
   # Allow --malloc-lib='' to override other settings
   [ -z  "$malloc_lib" ] && return
 
@@ -395,12 +477,11 @@ set_malloc_lib() {
       fi
       ;;
     *)
-      log_error "--malloc-lib must be an absolute path or 'tcmalloc'; " \
-        "ignoring value '$malloc_lib'"
+      log_error "--malloc-lib must be an absolute path, 'tcmalloc' or " \
+      "'jemalloc'; ignoring value '$malloc_lib'"
       exit 1
       ;;
   esac
-
   add_mysqld_ld_preload "$malloc_lib"
 }
 
@@ -447,7 +528,9 @@ fi
 
 helper=`find_in_bin mysqld_safe_helper`
 print_defaults=`find_in_bin my_print_defaults`
-
+# Check if helper exists
+command -v $helper --help >/dev/null 2>&1
+helper_exist=$?
 #
 # Second, try to find the data directory
 #
@@ -456,10 +539,6 @@ print_defaults=`find_in_bin my_print_defaults`
 if test -d $MY_BASEDIR_VERSION/data/mysql
 then
   DATADIR=$MY_BASEDIR_VERSION/data
-  if test -z "$defaults" -a -r "$DATADIR/my.cnf"
-  then
-    defaults="--defaults-extra-file=$DATADIR/my.cnf"
-  fi
 # Next try where the source installs put it
 elif test -d $MY_BASEDIR_VERSION/var/mysql
 then
@@ -471,24 +550,13 @@ fi
 
 if test -z "$MYSQL_HOME"
 then 
-  if test -r "$MY_BASEDIR_VERSION/my.cnf" && test -r "$DATADIR/my.cnf"
-  then
-    log_error "WARNING: Found two instances of my.cnf -
-$MY_BASEDIR_VERSION/my.cnf and
-$DATADIR/my.cnf
-IGNORING $DATADIR/my.cnf"
-
-    MYSQL_HOME=$MY_BASEDIR_VERSION
-  elif test -r "$DATADIR/my.cnf"
+  if test -r "$DATADIR/my.cnf"
   then
     log_error "WARNING: Found $DATADIR/my.cnf
-The data directory is a deprecated location for my.cnf, please move it to
+The data directory is not a valid location for my.cnf, please move it to
 $MY_BASEDIR_VERSION/my.cnf"
-    unsafe_my_cnf=1
-    MYSQL_HOME=$DATADIR
-  else
-    MYSQL_HOME=$MY_BASEDIR_VERSION
   fi
+  MYSQL_HOME=$MY_BASEDIR_VERSION
 fi
 export MYSQL_HOME
 
@@ -548,14 +616,18 @@ plugin_dir="${plugin_dir}${PLUGIN_VARIANT}"
 # Ensure that 'logger' exists, if it's requested
 if [ $want_syslog -eq 1 ]
 then
-  my_which logger > /dev/null 2>&1
-  if [ $? -ne 0 ]
+  if ! command -v logger > /dev/null
   then
     log_error "--syslog requested, but no 'logger' program found.  Please ensure that 'logger' is in your PATH, or do not specify the --syslog option to mysqld_safe."
     exit 1
   fi
 fi
 
+if [ $skip_err_log -eq 1 ]
+then
+  append_arg_to_args "--skip-log-error"
+fi
+
 if [ -n "$err_log" -o $want_syslog -eq 0 ]
 then
   if [ -n "$err_log" ]
@@ -622,6 +694,7 @@ then
   if test "$user" != "root" -o $SET_USER = 1
   then
     USER_OPTION="--user=$user"
+    GROUP_OPTION="--group=$group"
   fi
   if test -n "$open_files"
   then
@@ -637,14 +710,19 @@ fi
 safe_mysql_unix_port=${mysql_unix_port:-${MYSQL_UNIX_PORT:-@MYSQL_UNIX_ADDR@}}
 # Make sure that directory for $safe_mysql_unix_port exists
 mysql_unix_port_dir=`dirname $safe_mysql_unix_port`
-if [ ! -d $mysql_unix_port_dir ]
+if [ ! -d $mysql_unix_port_dir -a $dry_run -eq 0 ]
 then
-  if ! `mkdir -p $mysql_unix_port_dir`
+  if ! mkdir -p $mysql_unix_port_dir
   then
     log_error "Fatal error Can't create database directory '$mysql_unix_port'"
     exit 1
   fi
-  chown $user $mysql_unix_port_dir
+  if [ "$user" -a "$group" ]; then
+    chown $user:$group $mysql_unix_port_dir
+  else
+    [ "$user" ] && chown $user $mysql_unix_port_dir
+    [ "$group" ] && chgrp $group $mysql_unix_port_dir
+  fi
   chmod 755 $mysql_unix_port_dir
 fi
 
@@ -660,7 +738,7 @@ then
 does not exist or is not executable. Please cd to the mysql installation
 directory and restart this script from there as follows:
 ./bin/mysqld_safe&
-See http://dev.mysql.com/doc/mysql/en/mysqld-safe.html for more information"
+See https://mariadb.com/kb/en/mysqld_safe for more information"
   exit 1
 fi
 
@@ -746,7 +824,7 @@ fi
 #
 # If there exists an old pid file, check if the daemon is already running
 # Note: The switches to 'ps' may depend on your operating system
-if test -f "$pid_file"
+if test -f "$pid_file" && [ $dry_run -eq 0 ]
 then
   PID=`cat "$pid_file"`
   if @CHECK_PID@
@@ -775,7 +853,7 @@ fi
 if @TARGET_LINUX@ && test $flush_caches -eq 1
 then
   # Locate sync, ensure it exists.
-  if ! my_which sync > /dev/null 2>&1
+  if ! command -v sync > /dev/null
   then
     log_error "sync command not found, required for --flush-caches"
     exit 1
@@ -787,7 +865,7 @@ then
   fi
 
   # Locate sysctl, ensure it exists.
-  if ! my_which sysctl > /dev/null 2>&1
+  if ! command -v sysctl > /dev/null
   then
     log_error "sysctl command not found, required for --flush-caches"
     exit 1
@@ -822,6 +900,7 @@ fi
 #fi
 
 cmd="`mysqld_ld_preload_text`$NOHUP_NICENESS"
+[ $dry_run -eq 1 ] && cmd=''
 
 #
 # Set mysqld's memory interleave policy.
@@ -830,7 +909,7 @@ cmd="`mysqld_ld_preload_text`$NOHUP_NICENESS"
 if @TARGET_LINUX@ && test $numa_interleave -eq 1
 then
   # Locate numactl, ensure it exists.
-  if ! my_which numactl > /dev/null 2>&1
+  if ! command -v numactl > /dev/null
   then
     log_error "numactl command not found, required for --numa-interleave"
     exit 1
@@ -841,7 +920,7 @@ then
   fi
 
   # Launch mysqld with numactl.
-  cmd="$cmd numactl --interleave=all"
+  [ $dry_run -eq 0 ] && cmd="$cmd numactl --interleave=all"
 elif test $numa_interleave -eq 1
 then
   log_error "--numa-interleave is not supported on this platform"
@@ -854,9 +933,16 @@ do
   cmd="$cmd "`shell_quote_string "$i"`
 done
 cmd="$cmd $args"
+
+if [ $dry_run -eq 1 ]
+then
+  # RETURN or EXIT depending if the script is being sourced or not.
+  (return 2> /dev/null) && return || exit
+fi
+
+
 # Avoid 'nohup: ignoring input' warning
 test -n "$NOHUP_NICENESS" && cmd="$cmd < /dev/null"
-
 log_notice "Starting $MYSQLD daemon with databases from $DATADIR"
 
 # variable to track the current number of "fast" (a.k.a. subsecond) restarts
@@ -875,13 +961,34 @@ else
   exec 2>/dev/null
 fi
 
+# maximum number of wsrep restarts
+max_wsrep_restarts=0
+
 while true
 do
   rm -f "$pid_file"	# Some extra safety
 
   start_time=`date +%M%S`
 
-  eval_log_error "$cmd"
+  # Perform wsrep position recovery if wsrep_on=1, skip otherwise.
+  if test $wsrep_on -eq 1
+  then
+    # this sets wsrep_start_position_opt
+    wsrep_recover_position "$cmd"
+
+    [ $? -ne 0 ] && exit 1 #
+
+    [ -n "$wsrep_urls" ] && url=`wsrep_pick_url $wsrep_urls` # check connect address
+
+    if [ -z "$url" ]
+    then
+      eval_log_error "$cmd $wsrep_start_position_opt"
+    else
+      eval_log_error "$cmd $wsrep_start_position_opt --wsrep_cluster_address=$url"
+    fi
+  else
+    eval_log_error "$cmd"
+  fi
   end_time=`date +%M%S`
 
   if test ! -f "$pid_file"		# This is removed if normal shutdown
@@ -944,6 +1051,20 @@ do
       I=`expr $I + 1`
     done
   fi
+
+  if [ -n "$wsrep_restart" ]
+  then
+    if [ $wsrep_restart -le $max_wsrep_restarts ]
+    then
+      wsrep_restart=`expr $wsrep_restart + 1`
+      log_notice "WSREP: sleeping 15 seconds before restart"
+      sleep 15
+    else
+      log_notice "WSREP: not restarting wsrep node automatically"
+      break
+    fi
+  fi
+
   log_notice "mysqld restarted"
   if test -n "$crash_script"
   then
