@@ -14,19 +14,21 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
+#include "mariadb.h"
 #include "debug_sync.h"  // DEBUG_SYNC
 #include "table.h"       // TABLE, FOREIGN_KEY_INFO
 #include "sql_class.h"   // THD
 #include "sql_base.h"    // open_and_lock_tables
 #include "sql_table.h"   // write_bin_log
-#include "sql_handler.h" // mysql_ha_rm_tables
 #include "datadict.h"    // dd_recreate_table()
-#include "lock.h"        // MYSQL_OPEN_TEMPORARY_ONLY
+#include "lock.h"        // MYSQL_OPEN_* flags
 #include "sql_acl.h"     // DROP_ACL
 #include "sql_parse.h"   // check_one_table_access()
 #include "sql_truncate.h"
+#include "wsrep_mysqld.h"
 #include "sql_show.h"    //append_identifier()
-
+#include "sql_select.h"
+#include "sql_delete.h"
 
 /**
   Append a list of field names to a string.
@@ -38,15 +40,15 @@
 */
 
 static bool fk_info_append_fields(THD *thd, String *str,
-                                  List<LEX_STRING> *fields)
+                                  List<LEX_CSTRING> *fields)
 {
   bool res= FALSE;
-  LEX_STRING *field;
-  List_iterator_fast<LEX_STRING> it(*fields);
+  LEX_CSTRING *field;
+  List_iterator_fast<LEX_CSTRING> it(*fields);
 
   while ((field= it++))
   {
-    res|= append_identifier(thd, str, field->str, field->length);
+    res|= append_identifier(thd, str, field);
     res|= str->append(", ");
   }
 
@@ -78,22 +80,17 @@ static const char *fk_info_str(THD *thd, FOREIGN_KEY_INFO *fk_info)
     `db`.`tbl`, CONSTRAINT `id` FOREIGN KEY (`fk`) REFERENCES `db`.`tbl` (`fk`)
   */
 
-  res|= append_identifier(thd, &str, fk_info->foreign_db->str,
-                          fk_info->foreign_db->length);
+  res|= append_identifier(thd, &str, fk_info->foreign_db);
   res|= str.append(".");
-  res|= append_identifier(thd, &str, fk_info->foreign_table->str,
-                          fk_info->foreign_table->length);
+  res|= append_identifier(thd, &str, fk_info->foreign_table);
   res|= str.append(", CONSTRAINT ");
-  res|= append_identifier(thd, &str, fk_info->foreign_id->str,
-                          fk_info->foreign_id->length);
+  res|= append_identifier(thd, &str, fk_info->foreign_id);
   res|= str.append(" FOREIGN KEY (");
   res|= fk_info_append_fields(thd, &str, &fk_info->foreign_fields);
   res|= str.append(") REFERENCES ");
-  res|= append_identifier(thd, &str, fk_info->referenced_db->str,
-                          fk_info->referenced_db->length);
+  res|= append_identifier(thd, &str, fk_info->referenced_db);
   res|= str.append(".");
-  res|= append_identifier(thd, &str, fk_info->referenced_table->str,
-                          fk_info->referenced_table->length);
+  res|= append_identifier(thd, &str, fk_info->referenced_table);
   res|= str.append(" (");
   res|= fk_info_append_fields(thd, &str, &fk_info->referenced_fields);
   res|= str.append(')');
@@ -142,7 +139,7 @@ fk_truncate_illegal_if_parent(THD *thd, TABLE *table)
   table->file->get_parent_foreign_key_list(thd, &fk_list);
 
   /* Out of memory when building list. */
-  if (thd->is_error())
+  if (unlikely(thd->is_error()))
     return TRUE;
 
   it.init(fk_list);
@@ -150,14 +147,14 @@ fk_truncate_illegal_if_parent(THD *thd, TABLE *table)
   /* Loop over the set of foreign keys for which this table is a parent. */
   while ((fk_info= it++))
   {
-    if (my_strcasecmp(system_charset_info, fk_info->referenced_db->str,
-                      table->s->db.str) ||
-        my_strcasecmp(system_charset_info, fk_info->referenced_table->str,
-                      table->s->table_name.str) ||
-        my_strcasecmp(system_charset_info, fk_info->foreign_db->str,
-                      table->s->db.str) ||
-        my_strcasecmp(system_charset_info, fk_info->foreign_table->str,
-                      table->s->table_name.str))
+    if (lex_string_cmp(system_charset_info, fk_info->referenced_db,
+                       &table->s->db) ||
+        lex_string_cmp(system_charset_info, fk_info->referenced_table,
+                       &table->s->table_name) ||
+        lex_string_cmp(system_charset_info, fk_info->foreign_db,
+                       &table->s->db) ||
+        lex_string_cmp(system_charset_info, fk_info->foreign_table,
+                       &table->s->table_name))
       break;
   }
 
@@ -189,13 +186,13 @@ fk_truncate_illegal_if_parent(THD *thd, TABLE *table)
                         binlong the statement.
 */
 
-enum Truncate_statement::truncate_result
-Truncate_statement::handler_truncate(THD *thd, TABLE_LIST *table_ref,
-                                     bool is_tmp_table)
+enum Sql_cmd_truncate_table::truncate_result
+Sql_cmd_truncate_table::handler_truncate(THD *thd, TABLE_LIST *table_ref,
+                                         bool is_tmp_table)
 {
   int error= 0;
-  uint flags;
-  DBUG_ENTER("Truncate_statement::handler_truncate");
+  uint flags= 0;
+  DBUG_ENTER("Sql_cmd_truncate_table::handler_truncate");
 
   /*
     Can't recreate, the engine must mechanically delete all rows
@@ -203,9 +200,7 @@ Truncate_statement::handler_truncate(THD *thd, TABLE_LIST *table_ref,
   */
 
   /* If it is a temporary table, no need to take locks. */
-  if (is_tmp_table)
-    flags= MYSQL_OPEN_TEMPORARY_ONLY;
-  else
+  if (!is_tmp_table)
   {
     /* We don't need to load triggers. */
     DBUG_ASSERT(table_ref->trg_event_map == 0);
@@ -214,13 +209,13 @@ Truncate_statement::handler_truncate(THD *thd, TABLE_LIST *table_ref,
       or writing into the table. Yet, to open a write cursor we need
       a thr_lock lock. Allow to open base tables only.
     */
-    table_ref->required_type= FRMTYPE_TABLE;
+    table_ref->required_type= TABLE_TYPE_NORMAL;
     /*
       Ignore pending FLUSH TABLES since we don't want to release
       the MDL lock taken above and otherwise there is no way to
       wait for FLUSH TABLES in deadlock-free fashion.
     */
-    flags= MYSQL_OPEN_IGNORE_FLUSH | MYSQL_OPEN_SKIP_TEMPORARY;
+    flags= MYSQL_OPEN_IGNORE_FLUSH;
     /*
       Even though we have an MDL lock on the table here, we don't
       pass MYSQL_OPEN_HAS_MDL_LOCK to open_and_lock_tables
@@ -241,13 +236,13 @@ Truncate_statement::handler_truncate(THD *thd, TABLE_LIST *table_ref,
       DBUG_RETURN(TRUNCATE_FAILED_SKIP_BINLOG);
 
   error= table_ref->table->file->ha_truncate();
-  if (error)
+  if (unlikely(error))
   {
     table_ref->table->file->print_error(error, MYF(0));
     /*
       If truncate method is not implemented then we don't binlog the
-      statement. If truncation has failed in a transactional engine then also we
-      donot binlog the statment. Only in non transactional engine we binlog
+      statement. If truncation has failed in a transactional engine then also
+      we don't binlog the statment. Only in non transactional engine we binlog
       inspite of errors.
      */
     if (error == HA_ERR_WRONG_COMMAND ||
@@ -260,64 +255,6 @@ Truncate_statement::handler_truncate(THD *thd, TABLE_LIST *table_ref,
 }
 
 
-/*
-  Close and recreate a temporary table. In case of success,
-  write truncate statement into the binary log if in statement
-  mode.
-
-  @param  thd     Thread context.
-  @param  table   The temporary table.
-
-  @retval  FALSE  Success.
-  @retval  TRUE   Error.
-*/
-
-static bool recreate_temporary_table(THD *thd, TABLE *table)
-{
-  bool error= TRUE;
-  TABLE_SHARE *share= table->s;
-  HA_CREATE_INFO create_info;
-  handlerton *table_type= table->s->db_type();
-  DBUG_ENTER("recreate_temporary_table");
-
-  memset(&create_info, 0, sizeof(create_info));
-  create_info.options|= HA_LEX_CREATE_TMP_TABLE;
-
-  table->file->info(HA_STATUS_AUTO | HA_STATUS_NO_LOCK);
-
-  /*
-    If LOCK TABLES list is not empty and contains this table
-    then unlock the table and remove it from this list.
-  */
-  mysql_lock_remove(thd, thd->lock, table);
-
-  /* Don't free share. */
-  close_temporary_table(thd, table, FALSE, FALSE);
-
-  /*
-    We must use share->normalized_path.str since for temporary tables it
-    differs from what dd_recreate_table() would generate based
-    on table and schema names.
-  */
-  ha_create_table(thd, share->normalized_path.str, share->db.str,
-                  share->table_name.str, &create_info, 1);
-
-  if (open_table_uncached(thd, share->path.str, share->db.str,
-                          share->table_name.str, TRUE))
-  {
-    error= FALSE;
-    thd->thread_specific_used= TRUE;
-  }
-  else
-    rm_temporary_table(table_type, share->path.str);
-
-  free_table_share(share);
-  my_free(table);
-
-  DBUG_RETURN(error);
-}
-
-
 /*
   Handle locking a base table for truncate.
 
@@ -331,11 +268,11 @@ static bool recreate_temporary_table(THD *thd, TABLE *table)
   @retval  TRUE   Error.
 */
 
-bool Truncate_statement::lock_table(THD *thd, TABLE_LIST *table_ref,
-                                    bool *hton_can_recreate)
+bool Sql_cmd_truncate_table::lock_table(THD *thd, TABLE_LIST *table_ref,
+                                        bool *hton_can_recreate)
 {
   TABLE *table= NULL;
-  DBUG_ENTER("Truncate_statement::lock_table");
+  DBUG_ENTER("Sql_cmd_truncate_table::lock_table");
 
   /* Lock types are set in the parser. */
   DBUG_ASSERT(table_ref->lock_type == TL_WRITE);
@@ -357,26 +294,47 @@ bool Truncate_statement::lock_table(THD *thd, TABLE_LIST *table_ref,
   */
   if (thd->locked_tables_mode)
   {
-    if (!(table= find_table_for_mdl_upgrade(thd, table_ref->db,
-                                            table_ref->table_name, FALSE)))
+    if (!(table= find_table_for_mdl_upgrade(thd, table_ref->db.str,
+                                            table_ref->table_name.str, NULL)))
       DBUG_RETURN(TRUE);
 
-    *hton_can_recreate= ha_check_storage_engine_flag(table->s->db_type(),
+    *hton_can_recreate= ha_check_storage_engine_flag(table->file->ht,
                                                      HTON_CAN_RECREATE);
     table_ref->mdl_request.ticket= table->mdl_ticket;
   }
   else
   {
+    handlerton *hton;
+    bool is_sequence;
+
     /* Acquire an exclusive lock. */
     DBUG_ASSERT(table_ref->next_global == NULL);
     if (lock_table_names(thd, table_ref, NULL,
-                         thd->variables.lock_wait_timeout,
-                         MYSQL_OPEN_SKIP_TEMPORARY))
+                         thd->variables.lock_wait_timeout, 0))
       DBUG_RETURN(TRUE);
 
-    if (dd_check_storage_engine_flag(thd, table_ref->db, table_ref->table_name,
-                                     HTON_CAN_RECREATE, hton_can_recreate))
+    if (!ha_table_exists(thd, &table_ref->db, &table_ref->table_name,
+                         &hton, &is_sequence) ||
+        hton == view_pseudo_hton)
+    {
+      my_error(ER_NO_SUCH_TABLE, MYF(0), table_ref->db.str,
+               table_ref->table_name.str);
       DBUG_RETURN(TRUE);
+    }
+
+    if (!hton)
+    {
+      /*
+        The table exists, but its storage engine is unknown, perhaps not
+        loaded at the moment. We need to open and parse the frm to know the
+        storage engine in question, so let's proceed with the truncation and
+        try to open the table. This will produce the correct error message
+        about unknown engine.
+      */
+      *hton_can_recreate= false;
+    }
+    else
+      *hton_can_recreate= !is_sequence && hton->flags & HTON_CAN_RECREATE;
   }
 
   /*
@@ -388,19 +346,19 @@ bool Truncate_statement::lock_table(THD *thd, TABLE_LIST *table_ref,
   {
     DEBUG_SYNC(thd, "upgrade_lock_for_truncate");
     /* To remove the table from the cache we need an exclusive lock. */
-    if (wait_while_table_is_used(thd, table, HA_EXTRA_PREPARE_FOR_DROP,
-                                 TDC_RT_REMOVE_NOT_OWN_AND_MARK_NOT_USABLE))
+    if (wait_while_table_is_used(thd, table,
+          *hton_can_recreate ? HA_EXTRA_PREPARE_FOR_DROP : HA_EXTRA_NOT_USED))
       DBUG_RETURN(TRUE);
     m_ticket_downgrade= table->mdl_ticket;
     /* Close if table is going to be recreated. */
     if (*hton_can_recreate)
-      close_all_tables_for_name(thd, table->s, HA_EXTRA_NOT_USED);
+      close_all_tables_for_name(thd, table->s, HA_EXTRA_NOT_USED, NULL);
   }
   else
   {
     /* Table is already locked exclusively. Remove cached instances. */
-    tdc_remove_table(thd, TDC_RT_REMOVE_ALL, table_ref->db,
-                     table_ref->table_name, FALSE);
+    tdc_remove_table(thd, TDC_RT_REMOVE_ALL, table_ref->db.str,
+                     table_ref->table_name.str, FALSE);
   }
 
   DBUG_RETURN(FALSE);
@@ -421,43 +379,27 @@ bool Truncate_statement::lock_table(THD *thd, TABLE_LIST *table_ref,
   @retval  TRUE   Error.
 */
 
-bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
+bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
 {
   int error;
-  TABLE *table;
   bool binlog_stmt;
-  DBUG_ENTER("Truncate_statement::truncate_table");
+  DBUG_ENTER("Sql_cmd_truncate_table::truncate_table");
+
+  DBUG_ASSERT((!table_ref->table) ||
+              (table_ref->table && table_ref->table->s));
 
   /* Initialize, or reinitialize in case of reexecution (SP). */
   m_ticket_downgrade= NULL;
 
-  /* Remove table from the HANDLER's hash. */
-  mysql_ha_rm_tables(thd, table_ref);
-
   /* If it is a temporary table, no need to take locks. */
-  if ((table= find_temporary_table(thd, table_ref)))
+  if (is_temporary_table(table_ref))
   {
     /* In RBR, the statement is not binlogged if the table is temporary. */
     binlog_stmt= !thd->is_current_stmt_binlog_format_row();
 
-    /* Note that a temporary table cannot be partitioned. */
-    if (ha_check_storage_engine_flag(table->s->db_type(), HTON_CAN_RECREATE))
-    {
-      if ((error= recreate_temporary_table(thd, table)))
-        binlog_stmt= FALSE; /* No need to binlog failed truncate-by-recreate. */
+    thd->close_unused_temporary_table_instances(table_ref);
 
-      DBUG_ASSERT(! thd->transaction.stmt.modified_non_trans_table);
-    }
-    else
-    {
-      /*
-        The engine does not support truncate-by-recreate. Open the
-        table and invoke the handler truncate. In such a manner this
-        can in fact open several tables if it's a temporary MyISAMMRG
-        table.
-      */
-      error= handler_truncate(thd, table_ref, TRUE);
-    }
+    error= handler_truncate(thd, table_ref, TRUE);
 
     /*
       No need to invalidate the query cache, queries with temporary
@@ -470,6 +412,9 @@ bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
   {
     bool hton_can_recreate;
 
+    if (WSREP(thd) &&
+        wsrep_to_isolation_begin(thd, table_ref->db.str, table_ref->table_name.str, 0))
+        DBUG_RETURN(TRUE);
     if (lock_table(thd, table_ref, &hton_can_recreate))
       DBUG_RETURN(TRUE);
 
@@ -479,11 +424,13 @@ bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
         The storage engine can truncate the table by creating an
         empty table with the same structure.
       */
-      error= dd_recreate_table(thd, table_ref->db, table_ref->table_name);
+      error= dd_recreate_table(thd, table_ref->db.str, table_ref->table_name.str);
 
-      if (thd->locked_tables_mode && thd->locked_tables_list.reopen_tables(thd))
+      if (thd->locked_tables_mode && thd->locked_tables_list.reopen_tables(thd, false))
+      {
           thd->locked_tables_list.unlink_all_closed_tables(thd, NULL, 0);
-
+          error= 1;
+      }
       /* No need to binlog a failed truncate-by-recreate. */
       binlog_stmt= !error;
     }
@@ -495,13 +442,22 @@ bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
       */
       error= handler_truncate(thd, table_ref, FALSE);
 
+      if (error == TRUNCATE_OK && thd->locked_tables_mode &&
+          (table_ref->table->file->ht->flags &
+           HTON_REQUIRES_CLOSE_AFTER_TRUNCATE))
+      {
+        thd->locked_tables_list.mark_table_for_reopen(thd, table_ref->table);
+        if (unlikely(thd->locked_tables_list.reopen_tables(thd, true)))
+          thd->locked_tables_list.unlink_all_closed_tables(thd, NULL, 0);
+      }
+
       /*
         All effects of a TRUNCATE TABLE operation are committed even if
         truncation fails in the case of non transactional tables. Thus, the
         query must be written to the binary log. The only exception is a
         unimplemented truncate method.
       */
-      if (error == TRUNCATE_OK || error == TRUNCATE_FAILED_BUT_BINLOG)
+      if (unlikely(error == TRUNCATE_OK || error == TRUNCATE_FAILED_BUT_BINLOG))
         binlog_stmt= true;
       else
         binlog_stmt= false;
@@ -527,12 +483,11 @@ bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
     to a shared one.
   */
   if (m_ticket_downgrade)
-    m_ticket_downgrade->downgrade_exclusive_lock(MDL_SHARED_NO_READ_WRITE);
+    m_ticket_downgrade->downgrade_lock(MDL_SHARED_NO_READ_WRITE);
 
   DBUG_RETURN(error);
 }
 
-
 /**
   Execute a TRUNCATE statement at runtime.
 
@@ -541,18 +496,17 @@ bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
   @return FALSE on success.
 */
 
-bool Truncate_statement::execute(THD *thd)
+bool Sql_cmd_truncate_table::execute(THD *thd)
 {
   bool res= TRUE;
-  TABLE_LIST *first_table= thd->lex->select_lex.table_list.first;
-  DBUG_ENTER("Truncate_statement::execute");
+  TABLE_LIST *table= thd->lex->select_lex.table_list.first;
+  DBUG_ENTER("Sql_cmd_truncate_table::execute");
 
-  if (check_one_table_access(thd, DROP_ACL, first_table))
+  if (check_one_table_access(thd, DROP_ACL, table))
     DBUG_RETURN(res);
 
-  if (! (res= truncate_table(thd, first_table)))
+  if (! (res= truncate_table(thd, table)))
     my_ok(thd);
 
   DBUG_RETURN(res);
 }
-
