@@ -1,7 +1,7 @@
 /************* Tabext C++ Functions Source Code File (.CPP) ************/
-/*  Name: TABEXT.CPP  Version 1.0                                      */
+/*  Name: TABEXT.CPP  Version 1.1                                      */
 /*                                                                     */
-/*  (C) Copyright to the author Olivier BERTRAND          2017         */
+/*  (C) Copyright to the author Olivier BERTRAND          2017 - 2019  */
 /*                                                                     */
 /*  This file contains the TBX, TDB and OPJOIN classes functions.      */
 /***********************************************************************/
@@ -14,7 +14,7 @@
 #include "sql_class.h"
 #include "sql_servers.h"
 #include "sql_string.h"
-#if !defined(__WIN__)
+#if !defined(_WIN32)
 #include "osutil.h"
 #endif
 
@@ -142,8 +142,14 @@ bool EXTDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
 	Username = GetStringCatInfo(g, "User", NULL);
 	Password = GetStringCatInfo(g, "Password", NULL);
 
-	if ((Srcdef = GetStringCatInfo(g, "Srcdef", NULL)))
+	// Memory was Boolean, it is now integer
+	if (!(Memory = GetIntCatInfo("Memory", 0)))
+		Memory = GetBoolCatInfo("Memory", false) ? 1 : 0;
+
+	if ((Srcdef = GetStringCatInfo(g, "Srcdef", NULL))) {
 		Read_Only = true;
+		if (Memory == 2) Memory = 1;
+	}	// endif Srcdef
 
 	Qrystr = GetStringCatInfo(g, "Query_String", "?");
 	Sep = GetStringCatInfo(g, "Separator", NULL);
@@ -166,10 +172,6 @@ bool EXTDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
 	if (Catfunc == FNC_TABLE)
 		Tabtyp = GetStringCatInfo(g, "Tabtype", NULL);
 
-	// Memory was Boolean, it is now integer
-	if (!(Memory = GetIntCatInfo("Memory", 0)))
-		Memory = GetBoolCatInfo("Memory", false) ? 1 : 0;
-
 	Pseudo = 2;    // FILID is Ok but not ROWID
 	return false;
 } // end of DefineAM
@@ -284,6 +286,37 @@ int TDBEXT::Decode(PCSZ txt, char *buf, size_t n)
 	return 0;
 } // end of Decode
 
+/*
+  Count number of %s placeholders in string.
+  Returns -1 if other sprintf placeholders are found, .g %d
+*/
+static int count_placeholders(const char *fmt)
+{
+  int cnt= 0;
+  for (const char *p=fmt; *p; p++)
+  {
+    if (*p == '%')
+    {
+      switch (p[1])
+      {
+      case 's':
+        /* %s found */
+        cnt++;
+        p++;
+        break;
+      case '%':
+        /* masking char for % found */
+        p++;
+        break;
+      default:
+        /* some other placeholder found */
+        return -1;
+      }
+    }
+  }
+  return cnt;
+}
+
 /***********************************************************************/
 /*  MakeSrcdef: make the SQL statement from SRDEF option.              */
 /***********************************************************************/
@@ -308,16 +341,29 @@ bool TDBEXT::MakeSrcdef(PGLOBAL g)
 				? To_CondFil->Having : PlugDup(g, "1=1");
 		} // endif ph
 
-		if (!stricmp(ph, "W")) {
+		int n_placeholders = count_placeholders(Srcdef);
+		if (n_placeholders < 0)
+		{
+			strcpy(g->Message, "MakeSQL: Wrong place holders specification");
+			return true;
+		}
+
+		if (!stricmp(ph, "W") && n_placeholders <= 1) {
 			Query = new(g)STRING(g, strlen(Srcdef) + strlen(fil1));
 			Query->SetLength(sprintf(Query->GetStr(), Srcdef, fil1));
-		} else if (!stricmp(ph, "WH")) {
+		}
+		else if (!stricmp(ph, "WH") && n_placeholders <= 2)
+		{
 			Query = new(g)STRING(g, strlen(Srcdef) + strlen(fil1) + strlen(fil2));
 			Query->SetLength(sprintf(Query->GetStr(), Srcdef, fil1, fil2));
-		} else if (!stricmp(ph, "H")) {
+		}
+		else if (!stricmp(ph, "H") && n_placeholders <= 1)
+		{
 			Query = new(g)STRING(g, strlen(Srcdef) + strlen(fil2));
 			Query->SetLength(sprintf(Query->GetStr(), Srcdef, fil2));
-		} else if (!stricmp(ph, "HW")) {
+		}
+		else if (!stricmp(ph, "HW") && n_placeholders <= 2)
+		{
 			Query = new(g)STRING(g, strlen(Srcdef) + strlen(fil1) + strlen(fil2));
 			Query->SetLength(sprintf(Query->GetStr(), Srcdef, fil2, fil1));
 		} else {
@@ -445,6 +491,43 @@ bool TDBEXT::MakeSQL(PGLOBAL g, bool cnt)
 	return false;
 } // end of MakeSQL
 
+/***********************************************************************/
+/*  Remove the NAME_CONST functions that are added by procedures.      */
+/***********************************************************************/
+void TDBEXT::RemoveConst(PGLOBAL g, char *stmt)
+{
+	char *p, *p2;
+	char  val[1025], nval[1025];
+	int   n, nc;
+
+	while ((p = strstr(stmt, "NAME_CONST")))
+		if ((n = sscanf(p, "%*[^,],%1024[^)])%n", val, &nc))) {
+			if (trace(33))
+				htrc("p=%s\nn=%d val=%s nc=%d\n", p, n, val, nc);
+
+			*p = 0;
+
+			if ((p2 = strstr(val, "'"))) {
+				if ((n = sscanf(p2, "%*['\\]%1024[^'\\]", nval))) {
+					if (trace(33))
+						htrc("p2=%s\nn=%d nval=%s\n", p2, n, nval);
+
+					strcat(strcat(strcat(strcat(stmt, "'"), nval), "'"), p + nc);
+				} else
+					break;
+
+			} else
+				strcat(strcat(strcat(strcat(stmt, "("), val), ")"), p + nc);
+
+			if (trace(33))
+				htrc("stmt=%s\n", stmt);
+
+		} else
+			break;
+
+		return;
+} // end of RemoveConst
+
 /***********************************************************************/
 /*  MakeCommand: make the Update or Delete statement to send to the    */
 /*  MySQL server. Limited to remote values and filtering.              */
@@ -524,6 +607,8 @@ bool TDBEXT::MakeCommand(PGLOBAL g)
 			stmt[i++] = (Qrystr[k] == '`') ? q : Qrystr[k];
 		} while (Qrystr[k++]);
 
+		RemoveConst(g, stmt);
+
 		if (body)
 			strcat(stmt, body);
 
