@@ -1,4 +1,5 @@
 /* Copyright (C) 2006 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
+   Copyright (c) 2009, 2019, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -19,7 +20,10 @@
 #include "ma_sp_defs.h"
 #include "ma_rt_index.h"
 #include "ma_blockrec.h"
+#include "trnman.h"
+#include "ma_trnman.h"
 #include <m_ctype.h>
+#include "ma_crypt.h"
 
 #if defined(MSDOS) || defined(__WIN__)
 #ifdef __WIN__
@@ -78,6 +82,7 @@ MARIA_HA *_ma_test_if_reopen(const char *filename)
     mode	Mode of table (O_RDONLY | O_RDWR)
     data_file   Filedescriptor of data file to use < 0 if one should open
 	        open it.
+    internal_table <> 0 if this is an internal temporary table
 
  RETURN
     #   Maria handler
@@ -86,7 +91,8 @@ MARIA_HA *_ma_test_if_reopen(const char *filename)
 
 
 static MARIA_HA *maria_clone_internal(MARIA_SHARE *share,
-                                      int mode, File data_file)
+                                      int mode, File data_file,
+                                      uint internal_table)
 {
   int save_errno;
   uint errpos;
@@ -151,16 +157,16 @@ static MARIA_HA *maria_clone_internal(MARIA_SHARE *share,
     info.lock_type= F_WRLCK;
 
   _ma_set_data_pagecache_callbacks(&info.dfile, share);
-  bitmap_init(&info.changed_fields, changed_fields_bitmap,
-              share->base.fields, 0);
+  my_bitmap_init(&info.changed_fields, changed_fields_bitmap,
+                 share->base.fields, 0);
   if ((*share->init)(&info))
     goto err;
 
   /* The following should be big enough for all pinning purposes */
   if (my_init_dynamic_array(&info.pinned_pages,
                             sizeof(MARIA_PINNED_PAGE),
-                            max(share->base.blobs*2 + 4,
-                                MARIA_MAX_TREE_LEVELS*3), 16))
+                            MY_MAX(share->base.blobs*2 + 4,
+                                MARIA_MAX_TREE_LEVELS*3), 16, MYF(0)))
     goto err;
 
 
@@ -181,7 +187,7 @@ static MARIA_HA *maria_clone_internal(MARIA_SHARE *share,
   if (!share->base.born_transactional)   /* For transactional ones ... */
   {
     /* ... force crash if no trn given */
-    _ma_set_trn_for_table(&info, &dummy_transaction_object);
+    _ma_set_tmp_trn_for_table(&info, &dummy_transaction_object);
     info.state= &share->state.state;	/* Change global values by default */
   }
   else
@@ -207,10 +213,18 @@ static MARIA_HA *maria_clone_internal(MARIA_SHARE *share,
   if (share->options & HA_OPTION_TMP_TABLE)
     m_info->lock.type= TL_WRITE;
 
-  m_info->open_list.data= m_info->share_list.data= (void*) m_info;
-  maria_open_list=  list_add(maria_open_list,  &m_info->open_list);
-  share->open_list= list_add(share->open_list, &m_info->share_list);
-
+  if (!internal_table)
+  {
+    m_info->open_list.data= m_info->share_list.data= (void*) m_info;
+    maria_open_list=  list_add(maria_open_list,  &m_info->open_list);
+    share->open_list= list_add(share->open_list, &m_info->share_list);
+  }
+  else
+  {
+    /* We don't need to mark internal temporary tables as changed on disk */
+    share->internal_table= 1;
+    share->global_changed= 1;
+  }
   DBUG_RETURN(m_info);
 
 err:
@@ -236,20 +250,6 @@ err:
 } /* maria_clone_internal */
 
 
-/* Make a clone of a maria table */
-
-MARIA_HA *maria_clone(MARIA_SHARE *share, int mode)
-{
-  MARIA_HA *new_info;
-  mysql_mutex_lock(&THR_LOCK_maria);
-  new_info= maria_clone_internal(share, mode,
-                                 share->data_file_type == BLOCK_RECORD ?
-                                 share->bitmap.file.file : -1);
-  mysql_mutex_unlock(&THR_LOCK_maria);
-  return new_info;
-}
-
-
 /******************************************************************************
   open a MARIA table
 
@@ -263,12 +263,14 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
 {
   int kfile,open_mode,save_errno;
   uint i,j,len,errpos,head_length,base_pos,keys, realpath_err,
-    key_parts,unique_key_parts,fulltext_keys,uniques;
+    key_parts,base_key_parts,unique_key_parts,fulltext_keys,uniques;
+  uint internal_table= MY_TEST(open_flags & HA_OPEN_INTERNAL_TABLE);
+  uint file_version;
   size_t info_length;
   char name_buff[FN_REFLEN], org_name[FN_REFLEN], index_name[FN_REFLEN],
        data_name[FN_REFLEN];
-  uchar *disk_cache, *disk_pos, *end_pos;
-  MARIA_HA info,*m_info,*old_info;
+  uchar *UNINIT_VAR(disk_cache), *disk_pos, *end_pos;
+  MARIA_HA info, *UNINIT_VAR(m_info), *old_info;
   MARIA_SHARE share_buff,*share;
   double *rec_per_key_part;
   ulong  *nulls_per_key_part;
@@ -278,7 +280,6 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
   File data_file= -1;
   DBUG_ENTER("maria_open");
 
-  LINT_INIT(m_info);
   kfile= -1;
   errpos= 0;
   head_length=sizeof(share_buff.state.header);
@@ -287,6 +288,11 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
   realpath_err= my_realpath(name_buff, fn_format(org_name, name, "",
                                                  MARIA_NAME_IEXT,
                                                  MY_UNPACK_FILENAME),MYF(0));
+  if (realpath_err > 0) /* File not found, no point in looking further. */
+  {
+    DBUG_RETURN(NULL);
+  }
+
   if (my_is_symlink(org_name) &&
       (realpath_err || mysys_test_invalid_symlink(name_buff)))
   {
@@ -294,10 +300,11 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
     DBUG_RETURN(0);
   }
 
-  mysql_mutex_lock(&THR_LOCK_maria);
   old_info= 0;
+  if (!internal_table)
+    mysql_mutex_lock(&THR_LOCK_maria);
   if ((open_flags & HA_OPEN_COPY) ||
-      !(old_info=_ma_test_if_reopen(name_buff)))
+      (internal_table || !(old_info=_ma_test_if_reopen(name_buff))))
   {
     share= &share_buff;
     bzero((uchar*) &share_buff,sizeof(share_buff));
@@ -314,20 +321,20 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
                     });
     DEBUG_SYNC_C("mi_open_kfile");
     if ((kfile=mysql_file_open(key_file_kfile, name_buff,
-                               (open_mode=O_RDWR) | O_SHARE | O_NOFOLLOW,
+                               (open_mode=O_RDWR) | O_SHARE | O_NOFOLLOW | O_CLOEXEC,
                                MYF(MY_NOSYMLINKS))) < 0)
     {
       if ((errno != EROFS && errno != EACCES) ||
 	  mode != O_RDONLY ||
 	  (kfile=mysql_file_open(key_file_kfile, name_buff,
-                                 (open_mode=O_RDONLY) | O_SHARE | O_NOFOLLOW,
+                                 (open_mode=O_RDONLY) | O_SHARE | O_NOFOLLOW | O_CLOEXEC,
                                  MYF(MY_NOSYMLINKS))) < 0)
 	goto err;
     }
     share->mode=open_mode;
     errpos= 1;
-    if (mysql_file_pread(kfile,share->state.header.file_version, head_length, 0,
-                 MYF(MY_NABP)))
+    if (mysql_file_pread(kfile,share->state.header.file_version, head_length,
+                         0, MYF(MY_NABP)))
     {
       my_errno= HA_ERR_NOT_A_TABLE;
       goto err;
@@ -386,21 +393,11 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
       Allocate space for header information and for data that is too
       big to keep on stack
     */
-    if (!my_multi_malloc(MY_WME,
-                         &disk_cache, info_length+128,
-                         &rec_per_key_part,
-                         (sizeof(*rec_per_key_part) * HA_MAX_POSSIBLE_KEY *
-                          HA_MAX_KEY_SEG),
-                         &nulls_per_key_part,
-                         (sizeof(*nulls_per_key_part) * HA_MAX_POSSIBLE_KEY *
-                          HA_MAX_KEY_SEG),
-                         NullS))
+    if (!(disk_cache= my_malloc(info_length+128, MYF(MY_WME))))
     {
       my_errno=ENOMEM;
       goto err;
     }
-    share_buff.state.rec_per_key_part=   rec_per_key_part;
-    share_buff.state.nulls_per_key_part= nulls_per_key_part;
 
     end_pos=disk_cache+info_length;
     errpos= 3;
@@ -413,7 +410,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
     keys=    (uint) share->state.header.keys;
     uniques= (uint) share->state.header.uniques;
     fulltext_keys= (uint) share->state.header.fulltext_keys;
-    key_parts= mi_uint2korr(share->state.header.key_parts);
+    base_key_parts= key_parts= mi_uint2korr(share->state.header.key_parts);
     unique_key_parts= mi_uint2korr(share->state.header.unique_key_parts);
     if (len != MARIA_STATE_INFO_SIZE)
     {
@@ -423,7 +420,8 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
     }
     share->state_diff_length=len-MARIA_STATE_INFO_SIZE;
 
-    _ma_state_info_read(disk_cache, &share->state);
+    if (!_ma_state_info_read(disk_cache, &share->state))
+      goto err;
     len= mi_uint2korr(share->state.header.base_info_length);
     if (len != MARIA_BASE_INFO_SIZE)
     {
@@ -431,6 +429,14 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
 			    len,MARIA_BASE_INFO_SIZE));
     }
     disk_pos= _ma_base_info_read(disk_cache + base_pos, &share->base);
+    /*
+      Check if old version of Aria file. Version 0 has language
+      stored in header.not_used
+    */
+    file_version= (share->state.header.not_used == 0);
+    if (file_version == 0)
+      share->base.language= share->state.header.not_used;
+    
     share->state.state_length=base_pos;
     /* For newly opened tables we reset the error-has-been-printed flag */
     share->state.changed&= ~STATE_CRASHED_PRINTED;
@@ -539,7 +545,10 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
     set_if_smaller(max_data_file_length, INT_MAX32);
     set_if_smaller(max_key_file_length, INT_MAX32);
 #endif
-    share->base.max_data_file_length=(my_off_t) max_data_file_length;
+    /* For internal temporary tables, max_data_file_length is already set */
+    if (!internal_table || !share->base.max_data_file_length)
+      share->base.max_data_file_length=(my_off_t) max_data_file_length;
+    DBUG_ASSERT(share->base.max_data_file_length);
     share->base.max_key_file_length=(my_off_t) max_key_file_length;
 
     if (share->options & HA_OPTION_COMPRESS_RECORD)
@@ -553,9 +562,9 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
     share->open_file_name.length=   strlen(name);
     if (!my_multi_malloc(MY_WME,
 			 &share,sizeof(*share),
-			 &share->state.rec_per_key_part,
+			 &rec_per_key_part,
                          sizeof(double) * key_parts,
-                         &share->state.nulls_per_key_part,
+                         &nulls_per_key_part,
                          sizeof(long)* key_parts,
 			 &share->keyinfo,keys*sizeof(MARIA_KEYDEF),
 			 &share->uniqueinfo,uniques*sizeof(MARIA_UNIQUEDEF),
@@ -580,11 +589,16 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
       goto err;
     errpos= 4;
 
-    *share=share_buff;
-    memcpy((char*) share->state.rec_per_key_part,
-	   (char*) rec_per_key_part, sizeof(double)*key_parts);
-    memcpy((char*) share->state.nulls_per_key_part,
-	   (char*) nulls_per_key_part, sizeof(long)*key_parts);
+    *share= share_buff;
+    share->state.rec_per_key_part=   rec_per_key_part;
+    share->state.nulls_per_key_part= nulls_per_key_part;
+
+    memcpy((char*) rec_per_key_part,
+	   (char*) share_buff.state.rec_per_key_part,
+           sizeof(double)*base_key_parts);
+    memcpy((char*) nulls_per_key_part,
+	   (char*) share_buff.state.nulls_per_key_part,
+           sizeof(long)*base_key_parts);
     memcpy((char*) share->state.key_root,
 	   (char*) key_root, sizeof(my_off_t)*keys);
     strmov(share->unique_file_name.str, name_buff);
@@ -598,6 +612,12 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
                              LSN_STORE_SIZE + TRANSID_SIZE :
                              0) + KEYPAGE_KEYID_SIZE + KEYPAGE_FLAG_SIZE +
                             KEYPAGE_USED_SIZE);
+
+    if (MY_TEST(share->base.extra_options & MA_EXTRA_OPTIONS_ENCRYPTED))
+    {
+      share->keypage_header+= ma_crypt_get_index_page_header_space(share);
+    }
+
     {
       HA_KEYSEG *pos=share->keyparts;
       uint32 ftkey_nr= 1;
@@ -615,7 +635,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
         {
           /* Packed key, ensure we don't get overflow in underflow() */
           keyinfo->underflow_block_length=
-            max((int) (share->max_index_block_size - keyinfo->maxlength * 3),
+            MY_MAX((int) (share->max_index_block_size - keyinfo->maxlength * 3),
                 (int) (share->keypage_header + share->base.key_reflength));
           set_if_smaller(keyinfo->underflow_block_length,
                          keyinfo->block_length/3);
@@ -734,7 +754,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
                          share->base.is_nulls_extended +
                          share->base.null_bytes +
                          share->base.pack_bytes +
-                         test(share->options & HA_OPTION_CHECKSUM));
+                         MY_TEST(share->options & HA_OPTION_CHECKSUM));
     share->kfile.file= kfile;
 
     if (open_flags & HA_OPEN_COPY)
@@ -803,7 +823,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
       /* Need some extra bytes for decode_bytes */
       share->base.extra_rec_buff_size+= 7;
     }
-    share->base.default_rec_buff_size= max(share->base.pack_reclength +
+    share->base.default_rec_buff_size= MY_MAX(share->base.pack_reclength +
                                            share->base.extra_rec_buff_size,
                                            share->base.max_key_length);
 
@@ -831,6 +851,12 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
     disk_pos= _ma_column_nr_read(disk_pos, share->column_nr,
                                  share->base.fields);
 
+    if (MY_TEST(share->base.extra_options & MA_EXTRA_OPTIONS_ENCRYPTED))
+    {
+      if (!(disk_pos= ma_crypt_read(share, disk_pos)))
+        goto err;
+    }
+
     if ((share->data_file_type == BLOCK_RECORD ||
          share->data_file_type == COMPRESSED_RECORD))
     {
@@ -870,9 +896,17 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
 					(keys ? MARIA_INDEX_BLOCK_MARGIN *
 					 share->block_size * keys : 0));
     my_free(disk_cache);
+    my_free(share_buff.state.rec_per_key_part);
+    disk_cache= 0;
+    share_buff.state.rec_per_key_part= 0;
+
     _ma_setup_functions(share);
+    max_data_file_length= share->base.max_data_file_length;
     if ((*share->once_init)(share, info.dfile.file))
       goto err;
+    if (internal_table)
+      set_if_smaller(share->base.max_data_file_length,
+                     max_data_file_length);
     if (share->now_transactional)
     {
       /* Setup initial state that is visible for all */
@@ -1006,14 +1040,16 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags)
       data_file= share->bitmap.file.file;       /* Only opened once */
   }
 
-  if (!(m_info= maria_clone_internal(share, mode, data_file)))
+  if (!(m_info= maria_clone_internal(share, mode, data_file,
+                                     internal_table)))
     goto err;
 
   if (maria_is_crashed(m_info))
     DBUG_PRINT("warning", ("table is crashed: changed: %u",
                            share->state.changed));
 
-  mysql_mutex_unlock(&THR_LOCK_maria);
+  if (!internal_table)
+    mysql_mutex_unlock(&THR_LOCK_maria);
 
   m_info->open_flags= open_flags;
   DBUG_PRINT("exit", ("table: %p  name: %s",m_info, name));
@@ -1040,10 +1076,12 @@ err:
     (*share->once_end)(share);
     /* fall through */
   case 4:
+    ma_crypt_free(share);
     my_free(share);
     /* fall through */
   case 3:
     my_free(disk_cache);
+    my_free(share_buff.state.rec_per_key_part);
     /* fall through */
   case 1:
     mysql_file_close(kfile,MYF(0));
@@ -1052,7 +1090,8 @@ err:
   default:
     break;
   }
-  mysql_mutex_unlock(&THR_LOCK_maria);
+  if (!internal_table)
+    mysql_mutex_unlock(&THR_LOCK_maria);
   my_errno= save_errno;
   DBUG_RETURN (NULL);
 } /* maria_open */
@@ -1295,7 +1334,7 @@ static void setup_key_functions(register MARIA_KEYDEF *keyinfo)
 
 
 /**
-   @brief Function to save and store the header in the index file (.MYI)
+   @brief Function to save and store the header in the index file (.MAI)
 
    Operates under MARIA_SHARE::intern_lock if requested.
    Sets MARIA_SHARE::MARIA_STATE_INFO::is_of_horizon if transactional table.
@@ -1323,7 +1362,7 @@ uint _ma_state_info_write(MARIA_SHARE *share, uint pWrite)
 
   if (pWrite & MA_STATE_INFO_WRITE_LOCK)
     mysql_mutex_lock(&share->intern_lock);
-  else if (maria_multi_threaded)
+  else if (maria_multi_threaded && !share->temporary)
     mysql_mutex_assert_owner(&share->intern_lock);
   if (share->base.born_transactional && translog_status == TRANSLOG_OK &&
       !maria_in_recovery)
@@ -1334,7 +1373,7 @@ uint _ma_state_info_write(MARIA_SHARE *share, uint pWrite)
       is too new). Recovery does it by itself.
     */
     share->state.is_of_horizon= translog_get_horizon();
-    DBUG_PRINT("info", ("is_of_horizon set to LSN (%lu,0x%lx)",
+    DBUG_PRINT("info", ("is_of_horizon set to LSN " LSN_FMT "",
                         LSN_IN_PARTS(share->state.is_of_horizon)));
   }
   res= _ma_state_info_write_sub(share->kfile.file, &share->state, pWrite);
@@ -1458,6 +1497,16 @@ static uchar *_ma_state_info_read(uchar *ptr, MARIA_STATE_INFO *state)
   keys= (uint) state->header.keys;
   key_parts= mi_uint2korr(state->header.key_parts);
 
+  /* Allocate memory for key parts if not already done */
+  if (!state->rec_per_key_part &&
+      !my_multi_malloc(MY_WME,
+                       &state->rec_per_key_part,
+                       sizeof(*state->rec_per_key_part) * key_parts,
+                       &state->nulls_per_key_part,
+                       sizeof(*state->nulls_per_key_part) * key_parts,
+                       NullS))
+    DBUG_RETURN(0);
+
   state->open_count = mi_uint2korr(ptr);		ptr+= 2;
   state->changed= mi_uint2korr(ptr);			ptr+= 2;
   state->create_rename_lsn= lsn_korr(ptr);		ptr+= LSN_STORE_SIZE;
@@ -1567,7 +1616,7 @@ uint _ma_base_info_write(File file, MARIA_BASE_INFO *base)
   mi_int2store(ptr,base->null_bytes);                   ptr+= 2;
   mi_int2store(ptr,base->original_null_bytes);	        ptr+= 2;
   mi_int2store(ptr,base->field_offsets);	        ptr+= 2;
-  mi_int2store(ptr,0);				        ptr+= 2; /* reserved */
+  mi_int2store(ptr,base->language);		        ptr+= 2;
   mi_int2store(ptr,base->block_size);	        	ptr+= 2;
   *ptr++= base->rec_reflength;
   *ptr++= base->key_reflength;
@@ -1610,7 +1659,7 @@ static uchar *_ma_base_info_read(uchar *ptr, MARIA_BASE_INFO *base)
   base->null_bytes= mi_uint2korr(ptr);			ptr+= 2;
   base->original_null_bytes= mi_uint2korr(ptr);		ptr+= 2;
   base->field_offsets= mi_uint2korr(ptr);		ptr+= 2;
-                                                        ptr+= 2;
+  base->language= mi_uint2korr(ptr);		        ptr+= 2;
   base->block_size= mi_uint2korr(ptr);			ptr+= 2;
 
   base->rec_reflength= *ptr++;
@@ -1675,10 +1724,10 @@ my_bool _ma_keyseg_write(File file, const HA_KEYSEG *keyseg)
   ulong pos;
 
   *ptr++= keyseg->type;
-  *ptr++= keyseg->language;
+  *ptr++= keyseg->language & 0xFF; /* Collation ID, low byte */
   *ptr++= keyseg->null_bit;
   *ptr++= keyseg->bit_start;
-  *ptr++= keyseg->bit_end;
+  *ptr++= keyseg->language >> 8; /* Collation ID, high byte */
   *ptr++= keyseg->bit_length;
   mi_int2store(ptr,keyseg->flag);	ptr+= 2;
   mi_int2store(ptr,keyseg->length);	ptr+= 2;
@@ -1697,7 +1746,7 @@ uchar *_ma_keyseg_read(uchar *ptr, HA_KEYSEG *keyseg)
    keyseg->language	= *ptr++;
    keyseg->null_bit	= *ptr++;
    keyseg->bit_start	= *ptr++;
-   keyseg->bit_end	= *ptr++;
+   keyseg->language	+= ((uint16) (*ptr++)) << 8;
    keyseg->bit_length   = *ptr++;
    keyseg->flag		= mi_uint2korr(ptr);  ptr+= 2;
    keyseg->length	= mi_uint2korr(ptr);  ptr+= 2;
@@ -1818,24 +1867,31 @@ uchar *_ma_column_nr_read(uchar *ptr, uint16 *offsets, uint columns)
 void _ma_set_data_pagecache_callbacks(PAGECACHE_FILE *file,
                                       MARIA_SHARE *share)
 {
+  pagecache_file_set_null_hooks(file);
   file->callback_data= (uchar*) share;
   file->flush_log_callback= &maria_flush_log_for_page_none; /* Do nothing */
+  file->post_write_hook= maria_page_write_failure;
 
   if (share->temporary)
   {
-    file->read_callback=  &maria_page_crc_check_none;
-    file->write_callback= &maria_page_filler_set_none;
+    file->post_read_hook= &maria_page_crc_check_none;
+    file->pre_write_hook= &maria_page_filler_set_none;
   }
   else
   {
-    file->read_callback=  &maria_page_crc_check_data;
+    file->post_read_hook= &maria_page_crc_check_data;
     if (share->options & HA_OPTION_PAGE_CHECKSUM)
-      file->write_callback= &maria_page_crc_set_normal;
+      file->pre_write_hook= &maria_page_crc_set_normal;
     else
-      file->write_callback= &maria_page_filler_set_normal;
+      file->pre_write_hook= &maria_page_filler_set_normal;
     if (share->now_transactional)
       file->flush_log_callback= maria_flush_log_for_page;
   }
+
+  if (MY_TEST(share->base.extra_options & MA_EXTRA_OPTIONS_ENCRYPTED))
+  {
+    ma_crypt_set_data_pagecache_callbacks(file, share);
+  }
 }
 
 
@@ -1850,26 +1906,32 @@ void _ma_set_data_pagecache_callbacks(PAGECACHE_FILE *file,
 void _ma_set_index_pagecache_callbacks(PAGECACHE_FILE *file,
                                        MARIA_SHARE *share)
 {
+  pagecache_file_set_null_hooks(file);
   file->callback_data= (uchar*) share;
   file->flush_log_callback= &maria_flush_log_for_page_none; /* Do nothing */
-  file->write_fail= maria_page_write_failure;
+  file->post_write_hook= maria_page_write_failure;
 
   if (share->temporary)
   {
-    file->read_callback=  &maria_page_crc_check_none;
-    file->write_callback= &maria_page_filler_set_none;
+    file->post_read_hook= &maria_page_crc_check_none;
+    file->pre_write_hook= &maria_page_filler_set_none;
   }
   else
   {
-    file->read_callback=  &maria_page_crc_check_index;
+    file->post_read_hook=  &maria_page_crc_check_index;
     if (share->options & HA_OPTION_PAGE_CHECKSUM)
-      file->write_callback= &maria_page_crc_set_index;
+      file->pre_write_hook= &maria_page_crc_set_index;
     else
-      file->write_callback= &maria_page_filler_set_normal;
+      file->pre_write_hook= &maria_page_filler_set_normal;
 
     if (share->now_transactional)
       file->flush_log_callback= maria_flush_log_for_page;
   }
+
+  if (MY_TEST(share->base.extra_options & MA_EXTRA_OPTIONS_ENCRYPTED))
+  {
+    ma_crypt_set_index_pagecache_callbacks(file, share);
+  }
 }
 
 
@@ -1885,7 +1947,7 @@ int _ma_open_datafile(MARIA_HA *info, MARIA_SHARE *share)
   DEBUG_SYNC_C("mi_open_datafile");
   info->dfile.file= share->bitmap.file.file=
     mysql_file_open(key_file_dfile, share->data_file_name.str,
-                    share->mode | O_SHARE, MYF(flags));
+                    share->mode | O_SHARE | O_CLOEXEC, MYF(flags));
   return info->dfile.file >= 0 ? 0 : 1;
 }
 
@@ -1899,7 +1961,7 @@ int _ma_open_keyfile(MARIA_SHARE *share)
   mysql_mutex_lock(&share->intern_lock);
   share->kfile.file= mysql_file_open(key_file_kfile,
                                      share->unique_file_name.str,
-                                     share->mode | O_SHARE | O_NOFOLLOW,
+                                     share->mode | O_SHARE | O_NOFOLLOW | O_CLOEXEC,
                              MYF(MY_WME | MY_NOSYMLINKS));
   mysql_mutex_unlock(&share->intern_lock);
   return (share->kfile.file < 0);
